//: ©2016 John Haney, All Rights Reserved
//: 
//: [Previous](@previous)
//:
//: # Year Of Swift
//: ### February Coupling
//:
//: Now that we've learned about variables and functions, it's time to start putting things together a bit more, literally.
//:
//: First, let's revisit our grocery list. We had made a variable for the price of each item in the list, but instead we could collect up the items in an array. Let's make one now and then see what we can do with it.

let priceOfItems = [3.95, 2.95, 3.49]

//: We now have priceOfItems, which is an array of Doubles. You make an array by putting the items in between square brackets []. Just like if you were writing a list of things, make sure to put commas between each item in the list.
//: If you hold Option and click priceOfItems above, you will see the declaration:
//: 
//: let priceOfItems: [Double]
//: 
//: When you put the type of a variable, you do so after a colon in Swift. Until now, we have been letting the Playground decide what type our variables should be, but there will certainly be times where you need to put in the type yourself.
//:
//: Swift figured out that we make an array of all Double numbers, so it figured out the type for us automatically.
let priceOfOtherItems = [2, 3, 4]
//: Look at the type of priceOfOtherItems, it's [Int]. Swift didn't see any Double numbers (because everything was a whole number), so it made the array be an array of Int (whole numbers). Change one of the numbers above to 2.00 and then option click to see what type priceOfOtherItems is now.
let priceOfOtherItemsReally: [Double] = [2, 3, 4]
//: This one is an array of Double certainly, because we told Swift that's what we wanted.

//: Can I make an array of both Numbers and Strings?
let arrayOfThings: [Any] = [1, "2", 3.0]

//: But, should you?
//: 
//: No. No, you shouldn't.
//: 
//: So, you _can_ make an array of whatever you want if you make it of type [Any]. Any means seriously it could be anything. But, this is not really a good idea. One of the things that Swift does back is help you make sure you are using the right types of variables and gives you suggestions of how best to use them. By saying that things are Any, you are throwing away all of the great help that Swift is trying to give you. Seriously, please don't use Any unless you are really sure you have to.
//:
//: Ok, I'll step down off my soapbox. Back to arrays, let's see what we can do with them.
//: 
//: As a reminder, here's what we had for priceOfItems:
//: 
//: let priceOfItems = [3.95, 2.95, 3.49]

//: How many items?
priceOfItems.count

//: We can get the first and last items
priceOfItems.first
priceOfItems.last

//: We can loop thru arrays
for price in priceOfItems
{
    print("price: \(price)")
}
//: We just did something cool with the print() function here, we printed the value of the _price_ variable, by putting it inside one of these things: \()
//: 
//: It's important to use the backslash \ before the parentheses or you won't get what you want, see:
let somePrice = 6.99
"(somePrice)"
"/(somePrice)"
"\(somePrice)"

//: We can get one particular item by using square brackets after the array with a number for which item, like this:
priceOfItems[1]
priceOfItems.first
//: Hmm, that's not the first item in the list! I wanted the first item, but [1] gives me the second one. It turns out that computers don't start counting at 1, they start at zero. So, let's try that again...
priceOfItems[0]
priceOfItems.first
//: There we go.
priceOfItems[0]
priceOfItems[1]
priceOfItems[2]

//: The number inside the [] is called the index. With arrays, the items in them start at index zero and go up if the array has more items. What happens if you use an index that is too big? Try it by uncommenting this:
// priceOfItems[3]

//: Big red stop sign error! Yep, if you try to go past the end of an array in Swift, your app will stop. And crash. Bam, poof, it's gone.
//: 
//: How do you avoid doing that? Well, you can always go thru arrays using the for-in loop:
for price in priceOfItems
{
}
//: Or, if you need to pick a particular item but you're not sure if the index is not too big, you can test the index. Let's make a function that gives me the price of an item at a particular index in the list, and we'll make it safer by checking the index we get in.
func priceAtIndex(index: Int) -> Double
{
    if (index < priceOfItems.count)
    {
        return priceOfItems[index]
    }
    else
    {
        // I don't know, I guess it's free!
        return 0.00
    }
}

//: So we check if index is less than the number of items in priceOfItems. This works nicely because the index counting starts at zero, so the last index is _count - 1_, and we can test for < count. Now, what should we do when the index is too big? Well, I guessed that the item is free. I hope the manager of the store doesn't see this... 😀
//: 
//: Let's try it out:
priceAtIndex(0)
priceAtIndex(2)
priceAtIndex(3)

//: Ok, so now we made our priceAtIndex() in a way that's a little bit safe so we don't have a big crash. What a relief!
//: 
//: You noticed that I didn't sound so confident about our priceAtIndex() function, huh? Well, that's because there's still a problem here. And it's not part of the everything you need to know about in life that you learned from kindergarten. It's… negative numbers! Yep, Int can have negative numbers, so I can pass in negative numbers, and our brand new shiny function will crash too.  Try it if you like…
// priceAtIndex(-1)

//: Time to handle negative numbers too:
func priceAtIndexReally(index: Int) -> Double
{
    if index > 0
    {
        if index < priceOfItems.count
        {
            return priceOfItems[index]
        }
        else
        {
            // I don't know, I guess it's free!
            return 0.00
        }
    }
    else
    {
        // I don't know, I guess it's free!
        return 0.00
    }
}

// And, it works!
priceAtIndexReally(-1)

// Ok, so we fixed it. That's always good. But our function is getting a little messy. Now I have the return 0.00 in two places, and the two if statements. I could do this shorter if I could test for both things at the same time. Fortunately, I can! Unfortunately, it uses some weird symbols. I'm not sure why we have these symbols for this still, I guess because that's how my dad had to write code too, and it was good enough for him.
func priceAtIndexSeriously(index: Int) -> Double
{
    if (index > 0) && (index < priceOfItems.count)
    {
        return priceOfItems[index]
    }
    else
    {
        // I don't know, I guess it's free!
        return 0.00
    }
}

//: Ok, that looks a lot better. What do those && symbols do? They are a logical AND. Let's try it out:
true && true
false && true
true && false
false && false
//: && is true only when both the thing on the left and the thing on the right are true.
//:
//: The other big one you may use and see in other code is the logical OR.
//:
//: || is the logical OR, true if EITHER are true
true || true
false || true
true || false
false || false

//: Wow, we're covering all sorts of weird stuff now, when do we get to the coupling? I know, I know. Guess what, these operators are a way to couple two boolean values together, so there's your first bit of coupling. 😍
//:
//: Now, let's go back to our function, because I'd like to make another small change. Here's a better version I would write:
func priceAtIndexFinally(index: Int) -> Double
{
    if (index > 0) && (index < priceOfItems.count)
    {
        return priceOfItems[index]
    }

    // I don't know, I guess it's free!
    return 0.00
}

//: There you go, I got rid of the _else_ part, and I'm happy with this function.
//:
//: Ok, let's get back to our shopping list, remember:
//:
//: let priceOfItems = [3.95, 2.95, 3.49]
//:
//: Let's also include the names of these items, so that we can make a proper receipt.

let nameOfItems = ["Milk", "Cheese", "Ice Cream"]

//: Let's print out our receipt:
for name in nameOfItems
{
    print("\(name)")
}
for price in priceOfItems
{
    print("\(price)")
}

//: Well, that was ok, but we really want to print the name and the price on the same line. I did some looking and found a useful function for this coupling of two arrays. It's called zip() and here's how you can use it:
for (name, price) in zip(nameOfItems, priceOfItems)
{
    print("\(name): $\(price)")
}
//: That turned out much better, now we have the names and prices matched up together.  zip() is like a zipper, matching up each item in order in an array.
//:
//: So, zip() is nice for coupling two arrays together, but what if we add something else, like the barcode number?
let skusOfItems: [Int] = [449520, 293833, 494500]
//: zip() only takes two arrays, so this doesn't work:
// zip(nameOfItems, priceOfItems, skusOfItems)

//: we need a double-zipper, I guess…
for ((name, price), sku) in zip(zip(nameOfItems, priceOfItems), skusOfItems)
{
    print("\(name): $\(price)")
    print("   \(sku)")
}

// Well, how about a discount on the items?
let discountOfItems: [Double] = [0, 0, 0.99]

// Quadruple zip!
for ((name, price), (sku, discount)) in zip(zip(nameOfItems, priceOfItems), zip(skusOfItems, discountOfItems))
{
    print("\(name): $\(price)")
    print("   \(sku)   -$\(discount)")
}

// Well, this is getting silly. It's time for a better way to handle tracking more than a few things at once. Time for a Struct!

//: ## Summary
//: We just learned about:
//: * Arrays, a list of items you can use by index number
//: * Computers start counting at zero. The first array index is zero.
//: * You can put the value of a variable into a string by using \()
//: * zip() is a nice function to use to put together two arrays

//: [Next](@next)
