//: ©2016 John Haney, All Rights Reserved
//: 
//: [Previous](@previous)
//:
//: # Year Of Swift
//: ### Structs and Stuff
//:
//: We were making our grocery shopping cart receipt, and adding different values about each item, but it was getting a little crazy. Let's look at this another way.
//:
//: Really, we're trying to keep track of each item and think about the different things to do with each item. We can collect up the different information for each one using a struct. I'm going to call these items a Product.
struct Product
{
    let price: Double
    let name: String
    let sku: Int
    let discount: Double
}

//: A struct collects together many variables and can also collect together functions too, so you can have code that runs on your struct.
//:
//: What we just did up there is define a struct Product, and told Swift what types of things are in a Product. Here we went ahead and put in the four things we got to last.
//:
//: Great, now how do we make a Product?

let milk = Product(price: 3.95, name: "Milk", sku: 449520, discount: 0.00)
let cheese = Product(price: 2.95, name: "Cheese", sku: 293833, discount: 0.00)
let iceCream = Product(price: 3.49, name: "Ice Cream", sku: 494500, discount: 0.99)

//: Now, we can put them in an array
let products: [Product] = [milk, cheese, iceCream]

//: We can get to the product information by using the variable names after the item, let's look at each of the parts of one of our products:
milk.price
milk.name
milk.sku
milk.discount

//: Great, now let's make our simple receipt:
for product in products
{
    print("\(product.name): $\(product.price)")
}

//: And the one that needed the quadruple zip!
for product in products
{
    print("\(product.name): $\(product.price)")
    print("   \(product.sku): -$\(product.discount)")
}

//: Wow, we didn't have to look for that zip function because we put our variables together into the Product together where they belong nicely.
//:
//: Now that we have these great products, what if we could make a product catalog to keep all of my products, and I could look them up by that barcode sku thingy?  Well, of course we can. Let's make an array of my products.
let productCatalogList = [milk, cheese, iceCream]
//: Let's lookup a product in this list by it's sku:
for product in productCatalogList
{
    if product.sku == 293833
    {
        print("this one: \(product)")
    }
}
//: Ok, so we found the product we wanted, but this seems like something the computer could do in a much better way for us. Of course it can! Let's do it and then look at what we made.
let productCatalog: [Int: Product] = [milk.sku: milk, cheese.sku: cheese, iceCream.sku: iceCream]

//: productCatalog is a Dictionary. A dictionary in Swift (sometimes called a hash in other languages) is another collection of items. Where an array keeps a list of items in order and you lookup items in an arry by index number, with a dictionary, you lookup items in a dictionary based on a key. In an actual dictionary, you lookup definitions of words by the word. In the dictionary we just made, you lookup products by a number. You define a Dictionary type as [Key:Value]. Then we filled up the dictionary with our three products.
//:
//: Let's try out our productCatalog. To get values out, we use the similar square brackets after the variable name [], but this time, we are using the key instead of an index. So we pass in a SKU.
productCatalog[494500]
productCatalog[293833]

//: Great, let's go ahead and look at one of the products we got from our nifty catalog
let oneProduct = productCatalog[494500]
oneProduct?.name
//: What's that question mark doing there?
//:
//: Option click on oneProduct and see it's type:
//:
//: let oneProduct: Product?
//:
//: Another question mark. Seriously, what is the deal with these things?
//:
//: Product? is a different type than Product. Product? is an Optional that _might_ contain a Product. An Optional is a special type when you might have a real value, or you might not have any value at all. Think of an Optional as the box for a Product. It's a nice box and you know the shape of it and on the outside it tells you what's supposed to be inside the box. But maybe the box is empty.
//:
//: Let's look at a few things we can do with optionals:
oneProduct?.name
oneProduct!.name

if let theProduct = oneProduct
{
    theProduct.name
}
//: What did we do there, let's look at each one quickly:
//:
//: oneProduct?.name  -- the fancy name for this is conditional unwrapping the optional. We'll just call it unwrapping.
//:
//: oneProduct!.name  -- this is a forced unwrapping. You're telling the Playground that you are SURE that there is something in that optional, because if there isn't, your app will CRASH! Please, please stay away from using the ! You can almost always write your code another way to do the unwrapping with the safer ? instead.
//:
//: if let theProduct = oneProduct
//:
//: This is another way to unwrap the optional with an if-let. These are very useful in swift as you might often want to only run some code if the optional you get has a real value inside it. We'll see lots of this code in the future.
//:
//: Unlike an array, we are looking up by key value, so what happens when we try to use a key that isn't in the Dictionary? Uncomment this out and find out…
productCatalog[123456]
let noProduct = productCatalog[123456]

noProduct?.name
noProduct?.sku

//: We get the value nil when the key doesn't map to a value. So, if you're ever looking up a word in a real dictionary again, and don't find the word, then you can just say "nil." The nice thing about using the ? to unwrap
//:
//: Now that we have a catalog and we know how to deal with the Product? optional that it will return, we can write a scanner to give us the price for a product SKU.

func priceScanner(sku: Int) -> Double?
{
    if let product = productCatalog[sku]
    {
        return product.price
    }
    return nil
}

priceScanner(494500) // beep
priceScanner(293833) // beep
priceScanner(123456) // bonk!

//: ## Summary
//: We just learned about:
//: * Structs are a great way to collect different variables together
//: * Dictionaries are a collection to keep items where you can look them up by a key.
//: * Optionals are a wrapper around a real value, and can either be nil (no value) or have a real value inside them.
//: * Optional unwrapping by using ? or if-let

//: [Next](@next)
